#include <iostream>
#include <glut.h>
#include <stdint.h>
#include <list>
#include "glm/glm.hpp"
#include "glm/ext.hpp"

#include "ms3d/msModel.h"
#include "filesystem/disk_file.hpp"
#include "render/render.hpp"
#include "scene_graph/scene_graph.hpp"
#include "skeleton/skeleton.hpp"
using namespace t3d;

const uint32_t SemanticMvp = 0;
const uint32_t SemanticTex = 1;
const uint32_t SemanticJoints = 2;

const uint32_t ModelNodeType = 1;
const uint32_t AnimModelNodeType = 2;

struct Vertex{
    glm::vec3 position;
    glm::vec3 normal;
    glm::vec2 texcoord;
    glm::vec4 joint;
    glm::vec4 weight;
};

struct Color{
    Color(uint8_t r = 255, uint8_t g = 255, uint8_t b = 255, uint8_t a = 255)
        : r(r), g(g), b(b), a(a) {}
    uint8_t r, g, b, a;
};

class ExampleNode : public Node{
public:
    void setModelMatrix(const glm::mat4 &modelMatrix){
        m_modelMatrix = modelMatrix;
    }

    const glm::mat4 &modelMatrix() const{
        return m_modelMatrix;
    }

protected:
    glm::mat4 m_modelMatrix;
};

class ModelNode : public ExampleNode{
public:
    virtual uint32_t type() const{
        return ModelNodeType;
    }

    void setVbo(const std::shared_ptr<Vbo> &vbo){
        m_vbo = vbo;
    }

    void setTexture(const std::shared_ptr<Texture> &texture){
        m_texture = texture;
    }

    void setShader(const std::shared_ptr<Shader> &shader){
        m_shader = shader;
    }

    const std::shared_ptr<Vbo> vbo() const{
        return m_vbo;
    }

    const std::shared_ptr<Shader> shader() const{
        return m_shader;
    }

    const std::shared_ptr<Texture> texture() const{
        return m_texture;
    }

private:
    std::shared_ptr<Vbo> m_vbo;
    std::shared_ptr<Texture> m_texture;
    std::shared_ptr<Shader> m_shader;
};

class AnimModelNode : public ModelNode{
public:
    AnimModelNode(){
        m_builder = std::make_shared<SkeletonBuilder>();
    }

    virtual uint32_t type() const{
        return AnimModelNodeType;
    }

    const std::shared_ptr<Skeleton> skeleton() const{
        return m_skeleton;
    }

    const std::shared_ptr<SkeletonBuilder> builder() const{
        return m_builder;
    }

    void setSkeleton(const std::shared_ptr<Skeleton> &skeleton){
        m_skeleton = skeleton;
    }

    void setFrame(double frameTime){
        m_builder->buildSkeleton(m_skeleton, frameTime);
    }

private:
    std::shared_ptr<Skeleton> m_skeleton;
    std::shared_ptr<SkeletonBuilder> m_builder;
};

class ExampleRenderState : public RenderState{
public:
    ExampleRenderState()
        : projection(1.0), view(1.0), model(1.0)
    {}

    virtual RenderState *copy() const{
        ExampleRenderState *state = new ExampleRenderState();
        state->projection = projection;
        state->view = view;
        state->model = model;
        return state;
    }

    glm::mat4 projection, view, model;
};

class AnimModelNodeRenderDelegate : public NodeRenderDelegate{
public:
    virtual void render(const std::shared_ptr<Node> &node, const std::shared_ptr<RenderState> &renderState){
        std::shared_ptr<AnimModelNode> modelNode = std::static_pointer_cast<AnimModelNode>(node);
        std::shared_ptr<ExampleRenderState> exampleRenderState = std::static_pointer_cast<ExampleRenderState>(renderState);

        if (!modelNode->skeleton())
            return;

        static double frameTime = 0;
        frameTime += 0.2;
        modelNode->setFrame(frameTime);

        modelNode->vbo()->bind();
        modelNode->shader()->beginDraw();

        glm::mat4 mvp = exampleRenderState->projection * exampleRenderState->view * (exampleRenderState->model * modelNode->modelMatrix());
        modelNode->shader()->setUniformMatrix4fvArrayBySemantic(SemanticMvp, 1, &mvp[0][0]);
        modelNode->texture()->bind(0);
        modelNode->shader()->setUniform1iBySemantic(SemanticTex, 0);
        modelNode->shader()->setUniformMatrix4fvArrayBySemantic(SemanticJoints, modelNode->skeleton()->getCountJoints(), modelNode->builder()->getRawFinalMatrices());
        glDrawElements(GL_TRIANGLES, modelNode->vbo()->indicesSize() / 2, GL_UNSIGNED_SHORT, 0);

        modelNode->shader()->endDraw();
    }
};

class ModelNodeRenderDelegate : public NodeRenderDelegate{
public:
    virtual void render(const std::shared_ptr<Node> &node, const std::shared_ptr<RenderState> &renderState){
        std::shared_ptr<ModelNode> modelNode = std::static_pointer_cast<ModelNode>(node);
        std::shared_ptr<ExampleRenderState> exampleRenderState = std::static_pointer_cast<ExampleRenderState>(renderState);

        modelNode->vbo()->bind();
        modelNode->shader()->beginDraw();

        glm::mat4 mvp = exampleRenderState->projection * exampleRenderState->view * (exampleRenderState->model * modelNode->modelMatrix());
        modelNode->shader()->setUniformMatrix4fvArrayBySemantic(SemanticMvp, 1, &mvp[0][0]);
        modelNode->texture()->bind(0);
        modelNode->shader()->setUniform1iBySemantic(SemanticTex, 0);
        glDrawElements(GL_TRIANGLES, modelNode->vbo()->indicesSize() / 2, GL_UNSIGNED_SHORT, 0);

        modelNode->shader()->endDraw();
    }
};

class ExampleRenderDelegate : public RenderDelegate{
public:
    virtual std::shared_ptr<RenderState> createChildRenderState(const std::shared_ptr<Node> &node, const std::shared_ptr<RenderState> &renderState){
        std::shared_ptr<ExampleNode> exampleNode = std::static_pointer_cast<ExampleNode>(node);
        std::shared_ptr<ExampleRenderState> exampleRenderState = std::static_pointer_cast<ExampleRenderState>(renderState);
        std::shared_ptr<ExampleRenderState> resultRenderState(static_cast<ExampleRenderState*>(renderState->copy()));
        resultRenderState->model *= exampleNode->modelMatrix();
        return resultRenderState;
    }
};

std::shared_ptr<Scene> scene;
std::shared_ptr<ModelNode> model;
std::shared_ptr<ModelNode> loadModel(const std::shared_ptr<File> &file);

void init(){
    std::string vertexAnimSrc =
            "attribute vec4 position;\n"
            "attribute vec2 texcoord;\n"
            "attribute vec4 joint;\n"
            "attribute vec4 weight;\n"
            "uniform mat4 mvp;\n"
            "uniform mat4 joints[64];\n"
            "varying vec2 v_texcoord;\n"
            "void main(){\n"
            "   vec4 newVertex;\n"
            "   mat4 newMat = joints[int(joint.x)] * weight.x;\n"
            "   newMat += joints[int(joint.y)] * weight.y;\n"
            "   newMat += joints[int(joint.z)] * weight.z;\n"
            "   newMat += joints[int(joint.w)] * weight.w;\n"
            "   newVertex = newMat * position;\n"
            "   gl_Position = mvp * newVertex;\n"
            "   v_texcoord = texcoord;\n"
            "}\n";

    std::string vertexSrc =
            "attribute vec4 position;\n"
            "attribute vec2 texcoord;\n"
            "uniform mat4 mvp;\n"
            "varying vec2 v_texcoord;\n"
            "void main(){\n"
            "   gl_Position = mvp * position;\n"
            "   v_texcoord = texcoord;\n"
            "}\n";

    std::string fragmentSrc =
            "uniform sampler2D texture;\n"
            "varying vec2 v_texcoord;\n"
            "void main(){\n"
            "   gl_FragColor = texture2D(texture, v_texcoord);\n"
            "}\n";

    glEnable(GL_DEPTH_TEST);

    std::shared_ptr<Texture> texture = std::make_shared<Texture>();
    Color bitmap[32 * 32];
    for (uint y = 0; y < 32; ++y)
        for (uint x = 0; x < 32; ++x){
            bitmap[y * 32 + x] = Color(float(x) / 32.0f * 255, float(y) / 32.0f * 255, 1.0 * 255, 1.0 * 255);
        }
    texture->loadFromData(bitmap, 32, 32, 32);

    std::shared_ptr<Shader> shaderAnim = std::make_shared<Shader>();
    shaderAnim->loadFromSources(vertexAnimSrc, fragmentSrc);
    shaderAnim->appendAttribute(Attribute(shaderAnim->getAttributeHandle("position"), 3, 0, sizeof(Vertex)));
    shaderAnim->appendAttribute(Attribute(shaderAnim->getAttributeHandle("texcoord"), 2, 24, sizeof(Vertex)));
    shaderAnim->appendAttribute(Attribute(shaderAnim->getAttributeHandle("weight"), 4, 48, sizeof(Vertex)));
    shaderAnim->appendAttribute(Attribute(shaderAnim->getAttributeHandle("joint"), 4, 32, sizeof(Vertex)));
    shaderAnim->assignUniformSemanticByName("mvp", SemanticMvp);
    shaderAnim->assignUniformSemanticByName("joints", SemanticJoints);
    shaderAnim->assignUniformSemanticByName("texture", SemanticTex);

    std::shared_ptr<Shader> shader = std::make_shared<Shader>();
    shader->loadFromSources(vertexSrc, fragmentSrc);
    shader->appendAttribute(Attribute(shader->getAttributeHandle("position"), 3, 0, sizeof(Vertex)));
    shader->appendAttribute(Attribute(shader->getAttributeHandle("texcoord"), 2, 24, sizeof(Vertex)));
    shader->assignUniformSemanticByName("mvp", SemanticMvp);
    shader->assignUniformSemanticByName("texture", SemanticTex);

    std::shared_ptr<Vbo> vbo = std::make_shared<Vbo>();
    Vertex quadVertices[4];
    quadVertices[0].position = glm::vec3(-1, -1, 0); quadVertices[0].normal = glm::vec3(0, 0, -1); quadVertices[0].texcoord = glm::vec2(0, 0);
    quadVertices[1].position = glm::vec3(-1, 1, 0); quadVertices[1].normal = glm::vec3(0, 0, -1); quadVertices[1].texcoord = glm::vec2(0, 1);
    quadVertices[2].position = glm::vec3(1, 1, 0); quadVertices[2].normal = glm::vec3(0, 0, -1); quadVertices[2].texcoord = glm::vec2(1, 1);
    quadVertices[3].position = glm::vec3(1, -1, 0); quadVertices[3].normal = glm::vec3(0, 0, -1); quadVertices[3].texcoord = glm::vec2(1, 0);
    uint16_t quadIndices[6] = {0,1,2,0,2,3};
    vbo->loadFromData(&quadVertices[0], 4 * sizeof(Vertex), &quadIndices[0], 6 * sizeof(uint16_t));

    scene = std::make_shared<Scene>(std::make_shared<ExampleNode>());
    model = std::make_shared<ModelNode>();
    model->setShader(shader);
    model->setTexture(texture);
    model->setVbo(vbo);

    std::shared_ptr<ModelNode> model2 = loadModel(std::make_shared<DiskFile>(L"/Users/alexey/projects/tiny3d/dwarf.ms3d", FileOpenMode::Read));
    model2->setShader(shaderAnim);
    model2->setTexture(texture);
    model2->setModelMatrix(glm::translate(glm::vec3(2.0, 0.0, 0.0)) * glm::scale(0.01f, 0.01f, 0.01f));

    scene->rootSceneNode()->addNode(model);
    model->addNode(model2);

    std::shared_ptr<ExampleRenderDelegate> renderDelegate = std::make_shared<ExampleRenderDelegate>();
    std::shared_ptr<ModelNodeRenderDelegate> modelNodeRenderDelegate = std::make_shared<ModelNodeRenderDelegate>();
    std::shared_ptr<AnimModelNodeRenderDelegate> animModelNodeRenderDelegate = std::make_shared<AnimModelNodeRenderDelegate>();
    scene->assignRenderDelegate(ModelNodeType, modelNodeRenderDelegate);
    scene->assignRenderDelegate(AnimModelNodeType, animModelNodeRenderDelegate);
    scene->setRenderDelegate(renderDelegate);
}

void render(){
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    static float rot = 0;
    rot += 0.01;
    model->setModelMatrix(glm::eulerAngleYXZ(rot, 0.0f, 0.0f));
    std::shared_ptr<ExampleRenderState> renderState = std::make_shared<ExampleRenderState>();
    renderState->projection = glm::perspective(45.0f, 800.0f / 480.0f, 1.0f, 100.0f);
    renderState->view = glm::lookAt(glm::vec3(0, 0, -5), glm::vec3(0.0), glm::vec3(0, 1, 0));
    scene->render(renderState);

    glutSwapBuffers();
}

std::shared_ptr<ModelNode> loadModel(const std::shared_ptr<File> &file){
    uint32_t frameCount = 0;

    std::vector<Vertex> vertices;
    std::vector<uint16_t> indices;

    std::shared_ptr<AnimModelNode> model = std::make_shared<AnimModelNode>();

    msModel mm;
    mm.Load(file);
    mm.SetupJoints();
    mm.SetFrame(-1);

    uint32_t counter = 0;
    for (uint32_t k = 0; k < mm.GetNumGroups(); ++k){
        ms3d_group_t *group = mm.GetGroup(k);

//        if (group->materialIndex != -1){
//            try{
//                ms3d_material_t *material = mm.GetMaterial(group->materialIndex);

//                std::string texturePath = material->texture;
//                std::wstring wtexturePath = getFileName(strToWstr(texturePath), L'/');
//                wtexturePath = getFileName(wtexturePath, L'\\');

//                TexturePtr texture;
//                if(wtexturePath.size() > 0)
//                    texture = m_core->openTexture(wtexturePath);
//                smaterial->setTexture(texture, 0);
//            }catch(std::exception &){

//            }
//        }

        for (uint32_t i = 0; i < group->triangleIndices.size(); ++i){
            uint32_t tri = group->triangleIndices[i];
            ms3d_triangle_t *t = mm.GetTriangle(tri);

            ms3d_vertex_t* v[3];
            v[0] = mm.GetVertex(t->vertexIndices[0]);
            v[1] = mm.GetVertex(t->vertexIndices[1]);
            v[2] = mm.GetVertex(t->vertexIndices[2]);

            Vertex vert[3];
            for (int j = 0; j < 3; ++j){
                float w1 = v[j]->weights[0];
                float w2 = v[j]->weights[1];
                float w3 = v[j]->weights[2];
                float w4 = v[j]->weights[0] - v[j]->weights[1] - v[j]->weights[2];
                float k = 100.0f;
                w1 /= k;
                w2 /= k;
                w3 /= k;
                w4 /= k;

                if (w1 == 0)
                    w1 = 1.0;

                vert[j].position = glm::vec3(-v[j]->vertex[0], v[j]->vertex[1], v[j]->vertex[2]);
                vert[j].normal = glm::vec3(0, 0, 1);
                vert[j].texcoord = glm::vec2(t->s[j], 1.0-t->t[j]);
                vert[j].joint = glm::vec4(v[j]->boneId, v[j]->boneIds[0], v[j]->boneIds[1], v[j]->boneIds[2]);
                vert[j].weight = glm::vec4(w1, w2, w3, w4);
                indices.push_back(counter++);

                vertices.push_back(vert[j]);
            }
        }

        //grp
    }

    std::shared_ptr<Vbo> vbo = std::make_shared<Vbo>();
    vbo->loadFromData(&vertices[0], vertices.size() * sizeof(Vertex), &indices[0], indices.size() * sizeof(uint16_t));
    model->setVbo(vbo);

    std::shared_ptr<Skeleton> skeleton = std::make_shared<Skeleton>();

    for (uint32_t i = 0; i < mm.GetNumJoints(); ++i){
        ms3d_joint_t *joint = mm.GetJoint(i);
        BoneJoint bone;
        bone.setPosition(-joint->pos[0], joint->pos[1], joint->pos[2]);
        joint->rot[2] *= -1.0;
        joint->rot[1] *= -1.0;
        bone.bindRot = glm::quat_cast(glm::eulerAngleYXZ(joint->rot[1], joint->rot[0], joint->rot[2]));
        for (int j = 0; j < joint->rotationKeys.size(); ++j){
            ms3d_keyframe_t &posKey = joint->positionKeys[j];
            ms3d_keyframe_t &rotKey = joint->rotationKeys[j];
            BoneJointFrame frame;
            frame.setPosition(-posKey.key[0], posKey.key[1], posKey.key[2]);
            rotKey.key[2] *= -1.0;
            rotKey.key[1] *= -1.0;
            frame.rot = glm::quat_cast(glm::eulerAngleYXZ(rotKey.key[1], rotKey.key[0], rotKey.key[2]));
            frame.time = posKey.time;
            if (frameCount < frame.time)
                frameCount = frame.time;
            bone.frames.push_back(frame);
        }
        skeleton->addJoint(bone);
    }

    for (uint32_t i = 0; i < mm.GetNumJoints(); ++i){
        ms3d_joint_t *joint = mm.GetJoint(i);
        if (joint->parentIndex != -1){
            skeleton->getJoint(i).setParent(&(skeleton->getJoint(joint->parentIndex)));
        }
    }

//    for (uint32_t i = 0; i < model->getCountSurfaces(); ++i){
//        const SurfacePtr &surface = model->getSurface(i).surface;
        model->setSkeleton(skeleton);
//    }

    //model->setFrameCount(m_frameCount);
    return model;
}

void reshape(int w, int h){

}

int main(int argc, char **argv)
{
    glutInit(&argc, argv);
    glutInitWindowSize(800, 480);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
    glutCreateWindow("Test");

    init();

    glutDisplayFunc(render);
    glutReshapeFunc(reshape);
    glutIdleFunc(render);

    glutMainLoop();
    return 0;
}

