#ifndef RENDER_HPP
#define RENDER_HPP

#include <vector>

#ifdef __APPLE__
#include <OpenGL/gl.h>
#endif

namespace t3d{

const uint32_t SemanticsCount = 255;

class Attribute{
public:
    Attribute(GLint handle = -1, uint32_t size = 0, uint64_t offset = 0, uint32_t vertexSize = 0)
        : m_handle(handle), m_size(size), m_offset(offset), m_vertexSize(vertexSize) {}

    GLint handle() const{
        return m_handle;
    }

    uint32_t size() const{
        return m_size;
    }

    uint64_t offset() const{
        return m_offset;
    }

    uint32_t vertexSize() const{
        return m_vertexSize;
    }

private:
    GLint m_handle;
    uint32_t m_size;
    uint64_t m_offset;
    uint32_t m_vertexSize;
};

class Uniform{
public:
    Uniform(GLint handle = -1)
        : m_handle(handle) {}

    GLint handle() const{
        return m_handle;
    }

private:
    GLint m_handle;
};

class Vbo{
public:
    bool loadFromData(void *verticesData, uint64_t verticesSize, void *indicesData, uint64_t indicesSize){
        m_verticesSize = verticesSize;
        m_indicesSize = indicesSize;
        glGenBuffers(1, &m_vertex);
        glBindBuffer(GL_ARRAY_BUFFER, m_vertex);
        glBufferData(GL_ARRAY_BUFFER, verticesSize, verticesData, GL_STATIC_DRAW);
        glGenBuffers(1, &m_index);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_index);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, indicesSize, indicesData, GL_STATIC_DRAW);
        return true;
    }

    void bind(){
        glBindBuffer(GL_ARRAY_BUFFER, m_vertex);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m_index);
    }

    GLuint vertex() const{
        return m_vertex;
    }

    GLuint index() const{
        return m_index;
    }

    uint64_t verticesSize() const{
        return m_verticesSize;
    }

    uint64_t indicesSize() const{
        return m_indicesSize;
    }

    ~Vbo(){
        glDeleteBuffers(1, &m_vertex);
        glDeleteBuffers(1, &m_index);
    }

private:
    GLuint m_vertex;
    GLuint m_index;
    uint64_t m_verticesSize;
    uint64_t m_indicesSize;
};

class Shader{
public:
    bool loadFromSources(const std::string &vertexSource, const std::string &fragmentSource){
        m_vertex = glCreateShader(GL_VERTEX_SHADER);
        m_fragment = glCreateShader(GL_FRAGMENT_SHADER);
        const char *vertexSrc = vertexSource.c_str();
        const char *fragmentSrc = fragmentSource.c_str();
        int vertexSrcLen = strlen(vertexSrc);
        int fragmentSrcLen = strlen(fragmentSrc);
        glShaderSource(m_vertex, 1, &vertexSrc, &vertexSrcLen);
        glShaderSource(m_fragment, 1, &fragmentSrc, &fragmentSrcLen);
        glCompileShader(m_vertex);
        glCompileShader(m_fragment);
        m_program = glCreateProgram();
        glAttachShader(m_program, m_vertex);
        glAttachShader(m_program, m_fragment);
        glLinkProgram(m_program);
        return true;
    }

    GLuint getAttributeHandle(const std::string &name){
        return glGetAttribLocation(m_program, name.c_str());
    }

    GLuint getUniformHandle(const std::string &name){
        return glGetUniformLocation(m_program, name.c_str());
    }

    template <typename T>
    void beginDraw(T *pointer){
        bind();
        for (std::vector<Attribute>::iterator it = m_attributes.begin(), end = m_attributes.end(); it != end; ++it){
            glEnableVertexAttribArray(it->handle());
            glVertexAttribPointer(it->handle(), it->size(), GL_FLOAT, GL_FALSE, it->vertexSize(), pointer[it->offset()]);
        }
    }

    void beginDraw(){
        bind();
        for (std::vector<Attribute>::iterator it = m_attributes.begin(), end = m_attributes.end(); it != end; ++it){
            glEnableVertexAttribArray(it->handle());
            glVertexAttribPointer(it->handle(), it->size(), GL_FLOAT, GL_FALSE, it->vertexSize(), reinterpret_cast<void*>(it->offset()));
        }
    }

    void endDraw(){
        for (std::vector<Attribute>::iterator it = m_attributes.begin(), end = m_attributes.end(); it != end; ++it)
            glDisableVertexAttribArray(it->handle());
    }

    void bind(){
        glUseProgram(m_program);
    }

    GLuint program() const{
        return m_program;
    }

    GLuint vertex() const{
        return m_vertex;
    }

    GLuint fragment() const{
        return m_fragment;
    }

    void assignUniformSemantic(const Uniform &uniform, uint8_t semantic){
        m_uniformSemantics[semantic] = uniform;
    }

    void assignUniformSemanticByName(const std::string &name, uint8_t semantic){
        m_uniformSemantics[semantic] = getUniformHandle(name);
    }

    const Uniform &uniformBySemantic(uint8_t semantic) const{
        return m_uniformSemantics[semantic];
    }

    void appendAttribute(const Attribute &attribute){
        m_attributes.push_back(attribute);
    }

    const Attribute &attribute(uint8_t index) const{
        return m_attributes[index];
    }

    uint8_t countAttributes() const{
        return m_attributes.size();
    }

    void setUniform1iBySemantic(uint8_t semantic, int v0){
        glUniform1i(m_uniformSemantics[semantic].handle(), v0);
    }

    void setUniform2iBySemantic(uint8_t semantic, int v0, int v1){
        glUniform2i(m_uniformSemantics[semantic].handle(), v0, v1);
    }

    void setUniform3iBySemantic(uint8_t semantic, int v0, int v1, int v2){
        glUniform3i(m_uniformSemantics[semantic].handle(), v0, v1, v2);
    }

    void setUniform4iBySemantic(uint8_t semantic, int v0, int v1, int v2, int v3){
        glUniform4i(m_uniformSemantics[semantic].handle(), v0, v1, v2, v3);
    }

    void setUniform1fBySemantic(uint8_t semantic, float v0){
        glUniform1f(m_uniformSemantics[semantic].handle(), v0);
    }

    void setUniform2fBySemantic(uint8_t semantic, float v0, float v1){
        glUniform2f(m_uniformSemantics[semantic].handle(), v0, v1);
    }

    void setUniform3fBySemantic(uint8_t semantic, float v0, float v1, float v2){
        glUniform3f(m_uniformSemantics[semantic].handle(), v0, v1, v2);
    }

    void setUniform4fBySemantic(uint8_t semantic, float v0, float v1, float v2, float v3){
        glUniform4f(m_uniformSemantics[semantic].handle(), v0, v1, v2, v3);
    }

    void setUniform1ivBySemantic(uint8_t semantic, uint8_t count, int *array){
        glUniform1iv(m_uniformSemantics[semantic].handle(), count, array);
    }

    void setUniform2ivBySemantic(uint8_t semantic, uint8_t count, int *array){
        glUniform2iv(m_uniformSemantics[semantic].handle(), count, array);
    }

    void setUniform3ivBySemantic(uint8_t semantic, uint8_t count, int *array){
        glUniform3iv(m_uniformSemantics[semantic].handle(), count, array);
    }

    void setUniform4ivBySemantic(uint8_t semantic, uint8_t count, int *array){
        glUniform4iv(m_uniformSemantics[semantic].handle(), count, array);
    }

    void setUniform1fvBySemantic(uint8_t semantic, uint8_t count, float *array){
        glUniform1fv(m_uniformSemantics[semantic].handle(), count, array);
    }

    void setUniform2fvBySemantic(uint8_t semantic, uint8_t count, float *array){
        glUniform2fv(m_uniformSemantics[semantic].handle(), count, array);
    }

    void setUniform3fvBySemantic(uint8_t semantic, uint8_t count, float *array){
        glUniform3fv(m_uniformSemantics[semantic].handle(), count, array);
    }

    void setUniform4fvBySemantic(uint8_t semantic, uint8_t count, float *array){
        glUniform4fv(m_uniformSemantics[semantic].handle(), count, array);
    }

    void setUniformMatrix2fvArrayBySemantic(uint8_t semantic, uint8_t count, float *array){
        glUniformMatrix2fv(m_uniformSemantics[semantic].handle(), count, GL_FALSE, array);
    }

    void setUniformMatrix3fvArrayBySemantic(uint8_t semantic, uint8_t count, float *array){
        glUniformMatrix3fv(m_uniformSemantics[semantic].handle(), count, GL_FALSE, array);
    }

    void setUniformMatrix4fvArrayBySemantic(uint8_t semantic, uint8_t count, float *array){
        glUniformMatrix4fv(m_uniformSemantics[semantic].handle(), count, GL_FALSE, array);
    }

    ~Shader(){
        glDeleteShader(m_vertex);
        glDeleteShader(m_fragment);
        glDeleteProgram(m_program);
    }

private:
    GLuint m_program;
    GLuint m_vertex;
    GLuint m_fragment;
    Uniform m_uniformSemantics[SemanticsCount];
    std::vector<Attribute> m_attributes;
};

class Texture{
public:
    bool loadFromData(void *bitmapData, uint16_t width, uint16_t height, uint8_t bpp){
        glGenTextures(1, &m_handle);
        glBindTexture(GL_TEXTURE_2D, m_handle);
        GLenum type = GL_RGBA;
        if (bpp == 24) type = GL_RGB;
        glTexImage2D(GL_TEXTURE_2D, 0, type, width, height, 0, type, GL_UNSIGNED_BYTE, bitmapData);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        glGenerateMipmap(GL_TEXTURE_2D);
        return true;
    }

    void bind(uint8_t layer){
        glBindTexture(GL_TEXTURE_2D, m_handle);
        glActiveTexture(GL_TEXTURE0 + layer);
    }

    GLuint handle() const{
        return m_handle;
    }

    ~Texture(){
        glDeleteTextures(1, &m_handle);
    }

private:
    GLuint m_handle;
};

}

#endif
