#ifndef SCENE_GRAPH_HPP
#define SCENE_GRAPH_HPP

#include <vector>
#include <algorithm>
#include <memory>
#include <map>

namespace t3d{

//------------------------------------------------------------------------------------------------------------------------

class RenderState{
public:
    virtual RenderState *copy() const = 0;
    virtual ~RenderState() {}
};

//------------------------------------------------------------------------------------------------------------------------

class Node : public std::enable_shared_from_this<Node>{
public:
    Node()
        : m_type(0)
    {}

    std::shared_ptr<Node> parent() const{
        return m_parent.lock();
    }

    void setParent(const std::weak_ptr<Node> &parent){
        if (!m_parent.expired())
            m_parent.lock()->removeNode(shared_from_this());
        m_parent = parent;
        m_parent.lock()->m_children.push_back(shared_from_this());
    }

    bool addNode(const std::shared_ptr<Node> &node){
        if (!node->m_parent.expired())
            return false;
        m_children.push_back(node);
        node->m_parent = shared_from_this();
        return true;
    }

    void removeNode(const std::shared_ptr<Node> &node){
        node->m_parent.reset();
        m_children.erase(std::remove(m_children.begin(), m_children.end(), node), m_children.end());
    }

    uint32_t countNodes() const{
        return m_children.size();
    }

    const std::shared_ptr<Node> &node(uint32_t index) const{
        return m_children[index];
    }

    virtual uint32_t type() const{
        return m_type;
    }

protected:
    std::weak_ptr<Node> m_parent;
    std::vector<std::shared_ptr<Node>> m_children;
    uint32_t m_type;
};

//------------------------------------------------------------------------------------------------------------------------

class NodeRenderDelegate{
public:
    virtual void render(const std::shared_ptr<Node> &node, const std::shared_ptr<RenderState> &renderState) = 0;
    virtual ~NodeRenderDelegate() {}
};

//------------------------------------------------------------------------------------------------------------------------

class RenderDelegate{
public:
    virtual std::shared_ptr<RenderState> createChildRenderState(const std::shared_ptr<Node> &node, const std::shared_ptr<RenderState> &renderState) = 0;
    virtual ~RenderDelegate() {}
};

//------------------------------------------------------------------------------------------------------------------------

class Scene{
public:
    Scene(const std::shared_ptr<Node> &node)
        : m_rootSceneNode(node)
    {}

    const std::shared_ptr<Node> &rootSceneNode() const{
        return m_rootSceneNode;
    }

    void render(const std::shared_ptr<RenderState> &renderState){
        processNode(m_rootSceneNode, renderState);
    }

    void assignRenderDelegate(uint32_t type, const std::shared_ptr<NodeRenderDelegate> &renderDelegate){
        m_nodeRenderDelegates[type] = renderDelegate;
    }

    void setRenderDelegate(const std::shared_ptr<RenderDelegate> &renderDelegate){
        m_renderDelegate = renderDelegate;
    }

protected:
    void renderNode(const std::shared_ptr<Node> &node, const std::shared_ptr<RenderState> &renderState){
        uint32_t t = node->type();
        auto it = m_nodeRenderDelegates.find(t);
        if (it == m_nodeRenderDelegates.end())
            return;
        it->second->render(node, renderState);
    }

    void processNode(const std::shared_ptr<Node> &node, const std::shared_ptr<RenderState> &renderState){
       renderNode(node, renderState);
       std::shared_ptr<RenderState> subRenderState = m_renderDelegate->createChildRenderState(node, renderState);
       for (uint32_t i = 0; i < node->countNodes(); ++i){
           processNode(node->node(i), subRenderState);
       }
    }

    std::shared_ptr<Node> m_rootSceneNode;
    std::map<uint32_t, std::shared_ptr<NodeRenderDelegate>> m_nodeRenderDelegates;
    std::shared_ptr<RenderDelegate> m_renderDelegate;
};

}

#endif
