#pragma once

#include <stdint.h>
#include <vector>
#include "glm/glm.hpp"
#include "glm/ext.hpp"

namespace t3d{

struct BoneJointFrame{
    void setPosition(float x, float y, float z){
        pos[0] = x; pos[1] = y; pos[2] = z;
    }

    void setRotation(float x, float y, float z, float w){
        rot[0] = x; rot[1] = y; rot[2] = z; rot[3] = w;
    }

    glm::vec3 pos;
    glm::quat rot;
    int time;
};

struct BoneJoint{

    BoneJoint()
        : parent(0)
    {}

    // need sort frames by time for fast search
    void sortFrames(){
        qsort(frames.data(), frames.size(), sizeof(BoneJointFrame), [](const void *jf1, const void *jf2){
            return static_cast<const BoneJointFrame*>(jf1)->time - static_cast<const BoneJointFrame*>(jf2)->time;
        });
    }

    void setPosition(float x, float y, float z){
        bindPos[0] = x; bindPos[1] = y; bindPos[2] = z;
    }

    void setRotation(float x, float y, float z, float w){
        bindRot[0] = x; bindRot[1] = y; bindRot[2] = z; bindRot[3] = w;
    }

    void setParent(BoneJoint *parent){
        if (this->parent)
            this->parent->children.erase(std::remove(this->parent->children.begin(), this->parent->children.end(), this), this->parent->children.end());

        this->parent = parent;

        if (this->parent)
            this->parent->children.push_back(this);
    }

    BoneJoint *getParent() const{
        return this->parent;
    }

    std::vector<BoneJoint*> &getChildren(){
        return children;
    }

    // local bindPos/Rot matrix must be multiple by local pos/rot interpolated frame
    glm::vec3 bindPos;
    glm::quat bindRot;
    glm::mat4 localMat; // will calculate when build static data of skeleton (must build)
    glm::mat4 globInvMat; // will calculate when build static data of skeleton (must build)
    glm::mat4 globFrameMat; // time depends matrix
    std::vector<BoneJointFrame> frames;

private:
    BoneJoint *parent;
    std::vector<BoneJoint*> children;
};

class Skeleton{
public:
    Skeleton()
        : m_needRebuild(true)
    {}

    void addJoint(const BoneJoint &joint){
        m_joints.push_back(joint);
    }

    const BoneJoint &getJoint(uint32_t index) const{
        return m_joints[index];
    }

    // need manually set was built false
    BoneJoint &getJoint(uint32_t index){
        return m_joints[index];
    }

    uint32_t getCountJoints() const{
        return m_joints.size();
    }

    void clear(){
        m_joints.clear();
    }

    bool needRebuild() const{
        return m_needRebuild;
    }

    void setAsRebuilt(){
        m_needRebuild = false;
    }

    void setAsNeedRebuild(){
        m_needRebuild = true;
    }

private:
    std::vector<BoneJoint> m_joints;
    bool m_needRebuild;
};

class SkeletonBuilder{
public:
// skeleton builder
void buildSkeleton(const std::shared_ptr<Skeleton> &skeleton, double frameTime)
{
    m_finalMats.clear();

    if (skeleton->needRebuild()){
        // build local pose matrices
        for (uint32_t i = 0; i < skeleton->getCountJoints(); ++i){
            BoneJoint &joint = skeleton->getJoint(i);
            glm::mat4 poseMat = glm::mat4_cast(joint.bindRot);
            poseMat[3].x = joint.bindPos[0]; poseMat[3].y = joint.bindPos[1]; poseMat[3].z = joint.bindPos[2];
            joint.localMat = poseMat;
        }

        // calculate global matrices
        for (uint32_t i = 0; i < skeleton->getCountJoints(); ++i){
            BoneJoint &joint = skeleton->getJoint(i);
            glm::mat4 globInvMat = glm::inverse(calculateGlobalMatrix(joint));
            joint.globInvMat = globInvMat;
        }

        skeleton->setAsRebuilt();
    }

    // calculate frame matrices
    for (uint32_t i = 0; i < skeleton->getCountJoints(); ++i){
        BoneJoint &joint = skeleton->getJoint(i);
        if (!joint.getParent())
            calculateFramesMatrix(joint, glm::mat4(1.0), frameTime);
    }

    // calculate final matrix
    for (uint32_t i = 0; i < skeleton->getCountJoints(); ++i){
        BoneJoint &joint = skeleton->getJoint(i);
        m_finalMats.push_back(joint.globFrameMat * joint.globInvMat);
    }
}

void calculateFramesMatrix(BoneJoint &joint, const glm::mat4 &parentMatrix, double frameTime)
{
    int currentFrame = -1;
    int nextFrame = -1;

    // find current frame
    for (int i = joint.frames.size() - 1; i >= 0; --i)
        if (frameTime > joint.frames[i].time){
            currentFrame = i;
            break;
        }

    nextFrame = currentFrame + 1;

    if (currentFrame == -1)
        currentFrame = 0;

    if (nextFrame >= joint.frames.size())
        nextFrame = currentFrame;

    BoneJointFrame &frame1 = joint.frames[currentFrame];
    BoneJointFrame &frame2 = joint.frames[nextFrame];

    // get interpolation scale
    float a = frame2.time - frame1.time;
    float v = frameTime - frame1.time;
    float s = a != 0 ? v * 1.0 / a : 0;
    if (s < 0) s = 0;
    if (s > 1) s = 1;

    // build interpolated matrix
    glm::vec3 locFramePos = glm::lerp(frame1.pos, frame2.pos, s);
    glm::quat locFrameRot = glm::lerp(frame1.rot, frame2.rot, s);
    glm::mat4 locFrameMat = glm::mat4_cast(locFrameRot);
    locFrameMat[3].x = locFramePos[0]; locFrameMat[3].y = locFramePos[1]; locFrameMat[3].z = locFramePos[2];

    glm::mat4 result = parentMatrix * joint.localMat * locFrameMat;
    joint.globFrameMat = result;

    for (int i = 0; i < joint.getChildren().size(); ++i){
        BoneJoint *child = joint.getChildren().at(i);
        calculateFramesMatrix(*child, result, frameTime);
    }
}

glm::mat4 calculateGlobalMatrix(BoneJoint &joint)
{
    if (!joint.getParent())
        return joint.localMat;

    return calculateGlobalMatrix(*(joint.getParent())) * joint.localMat;
}

float *getRawFinalMatrices()
{
    // prepared for shader data
    return reinterpret_cast<float*>(m_finalMats.data());
}

private:
    std::vector<glm::mat4> m_finalMats;
};



}
